'''
@ summary:      Reads text files, return them as lists of text lines.
                Classes for model input (.gms-files) and outputs (.lst-files)
                Calculates F, G_y and J_uu.
@ author:       Axel Lodemel Holene
@ organization: Department of Chemical Engineering, NTNU, Norway
@ contact:      axel.holene@gmail.com
@ requires:     Python 2.7 or higher
@ todo:         Make "readFile" more efficient by using i.e. iteration with regex. 
'''

import re
import os



class AutoVivification(dict):
	"""Implementation of perl's autovivification feature."""
	def __getitem__(self, item):
		try:
			return dict.__getitem__(self, item)
		except KeyError:
			value = self[item] = type(self)()
			return value

class pf(float):
	''' Class for printing readable lists.'''
	def __repr__(self):
		return "%.10e" % self

class solvedModel:
    '''
    Class specifying the values retrieved from the .list file of the solved .gms model. 
    '''
    def __init__(self, filename):
        '''
        Assigns a name to the model (typically file name), then loads the values from "data".
        '''
        data = findData(readFile2(filename))
        self.name = filename
        self.inputfile = data.get('Inputfile')      # Sets name of model
        self.outputfile = data.get('Outputfile')
        self.FN2 = data.get('FN2')
        self.PumpWork = data.get('PumpWork')
        self.Z = data.get('Z')
        self.Tin = data.get('Tin')
        self.Tout = data.get('Tout')
        self.FCp = data.get('FCp')
        self.PresS1 = data.get('PresS1')
        self.TempS4 = data.get('TempS4')
        self.TempS5 = data.get('TempS5')
        self.TempS6 = data.get('TempS6')
        del data
    # def __init__(self, folder, filename):
    #     '''
    #     Assigns a name to the model (typically file name), then loads the values from "data".
    #     '''
    #     data = findData(readFile(folder, filename))
    #     self.name = filename
    #     self.inputfile = data.get('Inputfile')      # Sets name of model
    #     self.outputfile = data.get('Outputfile')
    #     self.FN2 = data.get('FN2')
    #     self.PumpWork = data.get('PumpWork')
    #     self.Z = data.get('Z')
    #     self.Tin = data.get('Tin')
    #     self.Tout = data.get('Tout')
    #     self.FCp = data.get('FCp')
    #     self.PresS1 = data.get('PresS1')
    #     self.TempS4 = data.get('TempS4')
    #     self.TempS5 = data.get('TempS5')
    #     self.TempS6 = data.get('TempS6')
    def __repr__(self):
        return "Class for solved gamsModel"

def readFile(folder,filename):
	try:
		with open(folder+'/'+filename, 'r') as f:
			res = f.read()
	except IOError:
		print "Error: Can't find file or read data named '%s'" % filename
	else:
		print "File named '%s' from the folder '%s' successfully loaded" % (filename, folder)
		f.close()
	return res

def readFile2(filename):
    try:
        with open(filename, 'r') as f:
            res = f.read()
    except IOError:
        print "Error: Can't find file or read data named '%s'" % filename
    else:
        print "File named '%s' successfully loaded" % (filename)
        f.close()
    return res

def findData(aString):
	oneVar = ['FN2', 'PumpWork', 'PresS1', 'TempS4', 'TempS5', 'TempS6', 'Z']
	multiVar = ['FCp', 'Tin']
	streams = ['H1', 'H2', 'H3', 'C1', 'C2']
	nameList = ['Input', 'Output']
	ans = AutoVivification() 
	for var in oneVar:
		try:
			ans[var] = re.match(r'^(.|\n)*?$\n.*VARIABLE '+var+'\.L\s*=\s*([0-9.E+-]+)', aString, re.MULTILINE).group(2)
		except:
			pass
	for var in multiVar:
		for stream in streams:
			try:
				ans[var][stream] = re.match(r'^(.|\n)*?$\n.*VARIABLE '+var+'\.L\s*\n.*\n.*'+stream+' ([0-9.E+-]+)', aString, re.MULTILINE).group(2)
			except:
				pass
	for aName in nameList:
		try:
			ans[aName+'file'] = re.match(r'^(.|\n)*?$\n.*'+aName+'\s+(\/\S+)+\/(\S+)$', aString, re.MULTILINE).group(3)
		except:
			pass
	return ans

def sens(aModelList, printing):
    '''
    aModel = solvedModel-object
    y = [FN2, PumpWork, pS1, TS4, TS5, TS6]
    d = [H1, ... , C2, FCp(H1), ... FCp(C2)]
    F = [[dFN2/dTin(H1), dFN2/dTin(H2), ... , dFN2/dFCp(H1), .... ],
        [dPW/dTin(H1), ... , ]]
    finds dFN2 and dPW for each model in the list
    computes dTin(H1) etc. Disregards current iteration and continues
    looping if dTin(H1) is zero. If it's not, continuing to FCp, 
    repeating. 
    '''
    y = ['FN2', 'PumpWork', 'PresS1', 'TempS4', 'TempS5', 'TempS6']
    subs = ['Tin', 'FCp']
    streams = ['H1', 'H2', 'H3', 'C1', 'C2']
    F = [[] for _ in range(len(y))]
    for aModel in aModelList:
        for y_i in y:
            if printing: print "y_i = %s" % y_i
            dy = float(getattr(aModel,y_i)) - float(getattr(ref,y_i))                #Finding d(y_i)
            for sub in subs:
                if printing: print "sub = %s" % sub
                for stream in streams:                                               #finding d(u_i)
                    if printing: print "stream = %s" % stream
                    dd = float(getattr(aModel,sub)[stream]) - \
                         float(getattr(ref,sub)[stream])
                    if dd != 0:                                                    #if it's not zero
                        F[y.index(y_i)].append(dy/dd)                           #dy/du is calculated
                        break                                                  #skipping to next u_i
                    else:                                                         #else, we continue
                        continue
    return F

def gain(aModelList):
    u = ['FN2', 'PumpWork']
    y = ['FN2', 'PumpWork', 'PresS1', 'TempS4', 'TempS5', 'TempS6']
    g = [[0 for _ in range(len(y))] for _ in range(len(u))]
    for (u_i, aModel) in zip(u,aModelList):
        du = float(getattr(aModel,u_i)) - float(getattr(ref,u_i))
        if printing: print "du = %s = %s - %s" % (u_i, getattr(aModel,u_i), getattr(ref,u_i))
        for y_i in y:
            g[u.index(u_i)][y.index(y_i)] = (float(getattr(aModel,y_i)) - float(getattr(ref,y_i)))/du
    return g

def nonDiagH(aModelList, printing):
    '''
    nondiagonal: h12, h21
    Computing d^2/(dx_i dx_j) by central differences.
    aModelList is a two-dimensional list of solvedModel objects with pp, pm, mp and mm perturbations
    aModelList = [fpp, fpm, fmp, fmm]
    fpp = f((hx)_i+(hx)_j), fpm = f((hx)_i-(hx)_j), etc
    '''
    eps = 1.e-8     # numerical tolerance
    if printing: print "nonDiagH runs: ****************************************"
    hi = abs(float(aModelList[0].FN2) - float(ref.FN2))
    hj = abs(float(aModelList[0].PumpWork) - float(ref.PumpWork))
    if hi == 0:
        step = 4*hj**2
        if printing: print "hi = 0, step = %4.4e" % step
    elif hj == 0:
        step = 4*hi**2
        if printing: print "hj = 0, step = %4.4e" % step
    else:
        step = 4*hi*hj
    if printing: print "step = %4.4e" % step
    if printing: print "Z values:\n %.8e, %.8e, %.8e, %.8e" % (float(aModelList[0].Z),float(aModelList[1].Z),float(aModelList[2].Z),float(aModelList[3].Z))
    if printing: print "objective = \n %.8e" % (float(aModelList[0].Z) - float(aModelList[1].Z) - float(aModelList[2].Z) + float(aModelList[3].Z))
    if printing: print "nonDiagH finished: ************************************"
    return (float(aModelList[0].Z) - float(aModelList[1].Z) - float(aModelList[2].Z) + float(aModelList[3].Z))/step

def diagH(aModelList, printing):
    eps = 1.e-8     # numerical tolerance
    hi = abs(float(aModelList[4].FN2) - float(ref.FN2))
    hj = abs(float(aModelList[4].PumpWork) - float(ref.PumpWork))
    if printing: print "diagH runs: ****************************************"
    if hi < eps:
        step = 12*hj**2
        if printing: print "hi = 0, step = %4.4e" % step
    elif hj < eps:
        step = 12*hi**2
        if printing: print "hj = 0, step = %4.4e" % step
    else:
        step = -1
        if printing: print "h1 and h2 < eps, step = -1"
    z_pp = float(aModelList[0].Z); z_pm = float(aModelList[1].Z)
    z = float(aModelList[2].Z); z_mm = float(aModelList[3].Z)
    z_p = float(aModelList[4].Z); z_m = float(aModelList[5].Z)
    # z = float(ref.Z)
    if printing: print "Z values:\n %.8e, %.8e, %.8e, %.8e, %.8e" % (z_pp, z_p, z, z_m, z_mm)
    if printing: print "objective = \n %.8e" % (-1*z_pp + 16*z_p - 30*z + 16*z_m - z_mm)
    if printing: print "diagH finished: ************************************"
    return (-1*z_pp + 16*z_p - 30*z + 16*z_m - z_mm)/step

def Hessian(aModelList, printing):
    '''
    aModelList = [hlist11, hlist12, hlist21, hlist22]
    '''
    return [[diagH(aModelList[0], printing), nonDiagH(aModelList[1], printing)], \
           [ nonDiagH(aModelList[2], printing), diagH(aModelList[3], printing) ] ]

def printMatrix(nestedList):
    try:
        print map(pf,nestedList)
    except:
        for line in nestedList:
            print map(pf,line)



def sortPrintDict(mydict):
    for key in sorted(mydict.iterkeys()):
        print "%s: %s" % (key, mydict[key])

# Global reference object
ref = solvedModel('mhex_ref.lst')

sensitivity = [\
  solvedModel('mhex_sens_H1.lst'), \
  solvedModel('mhex_sens_H2.lst'), \
  solvedModel('mhex_sens_H3.lst'), \
  solvedModel('mhex_sens_C1.lst'), \
  solvedModel('mhex_sens_C2.lst'),\
  solvedModel('sens_FCP_H1.lst'),\
  solvedModel('sens_FCP_H2.lst'),\
  solvedModel('sens_FCP_H3.lst'),\
  solvedModel('sens_FCP_C1.lst'),\
  solvedModel('sens_FCP_C2.lst')\
  ]

gainModels = \
    [solvedModel('gain_FN2.lst'),  
     solvedModel('gain_PW.lst')]

hessianModels = [\
	[solvedModel('h_11_hessian_pp.lst'),   \
	 solvedModel('h_11_hessian_pm.lst'),   \
	 solvedModel('h_11_hessian_mp.lst'),   \
	 solvedModel('h_11_hessian_mm.lst'),   \
     solvedModel('h_11_hessian_p.lst'),    \
     solvedModel('h_11_hessian_m.lst')],   \
	[solvedModel('h_12_hessian_pp.lst'),   \
	 solvedModel('h_12_hessian_pm.lst'),   \
	 solvedModel('h_12_hessian_mp.lst'),   \
	 solvedModel('h_12_hessian_mm.lst')],  \
	[solvedModel('h_21_hessian_pp.lst'),   \
	 solvedModel('h_21_hessian_pm.lst'),   \
	 solvedModel('h_21_hessian_mp.lst'),   \
	 solvedModel('h_21_hessian_mm.lst')],  \
	[solvedModel('h_22_hessian_pp.lst'),   \
	 solvedModel('h_22_hessian_pm.lst'),   \
	 solvedModel('h_22_hessian_mp.lst'),   \
	 solvedModel('h_22_hessian_mm.lst'),   \
     solvedModel('h_22_hessian_p.lst'),    \
     solvedModel('h_22_hessian_m.lst')]    \
	]

printing = True

F = sens(sensitivity, printing)
print "F = \n", printMatrix(F)
Gy = gain(gainModels)
print "Gy = \n", printMatrix(Gy)
Juu = Hessian(hessianModels, printing)
print "Juu =\n", printMatrix(Juu)

i = 1
j = 1
for aModelList in hessianModels:
    for aModel in aModelList:
        print i,j,aModel.Z
        j += 1
    j = 1
    i += 1

print "ref.z = ", ref.Z

# ref = solvedModel('ref', 'mhex_ref.lst')

# sensitivity = [\
#   solvedModel('sensitivity', 'mhex_sens_H1.lst'), \
#   solvedModel('sensitivity', 'mhex_sens_H2.lst'), \
#   solvedModel('sensitivity', 'mhex_sens_H3.lst'), \
#   solvedModel('sensitivity', 'mhex_sens_C1.lst'), \
#   solvedModel('sensitivity', 'mhex_sens_C2.lst'),\
#   solvedModel('sensitivity', 'sens_FCP_H1.lst'),\
#   solvedModel('sensitivity', 'sens_FCP_H2.lst'),\
#   solvedModel('sensitivity', 'sens_FCP_H3.lst'),\
#   solvedModel('sensitivity', 'sens_FCP_C1.lst'),\
#   solvedModel('sensitivity', 'sens_FCP_C2.lst')\
#   ]

# gainModels = \
#     [solvedModel('gain', 'gain_FN2.lst'),  
#      solvedModel('gain', 'gain_PW.lst')]

# hessianModels = [\
#     [solvedModel('hessian/h_11', 'hessian_pp.lst'), \
#      solvedModel('hessian/h_11', 'hessian_pm.lst'), \
#      solvedModel('hessian/h_11', 'hessian_mp.lst'), \
#      solvedModel('hessian/h_11', 'hessian_mm.lst'), \
#      solvedModel('hessian/h_11', 'hessian_p.lst'),  \
#      solvedModel('hessian/h_11', 'hessian_m.lst')], \
#     [solvedModel('hessian/h_12', 'hessian_pp.lst'), \
#      solvedModel('hessian/h_12', 'hessian_pm.lst'), \
#      solvedModel('hessian/h_12', 'hessian_mp.lst'), \
#      solvedModel('hessian/h_12', 'hessian_mm.lst')],\
#     [solvedModel('hessian/h_21', 'hessian_pp.lst'), \
#      solvedModel('hessian/h_21', 'hessian_pm.lst'), \
#      solvedModel('hessian/h_21', 'hessian_mp.lst'), \
#      solvedModel('hessian/h_21', 'hessian_mm.lst')],\
#     [solvedModel('hessian/h_22', 'hessian_pp.lst'), \
#      solvedModel('hessian/h_22', 'hessian_pm.lst'), \
#      solvedModel('hessian/h_22', 'hessian_mp.lst'), \
#      solvedModel('hessian/h_22', 'hessian_mm.lst'), \
#      solvedModel('hessian/h_22', 'hessian_p.lst'),  \
#      solvedModel('hessian/h_22', 'hessian_m.lst')] \
#     ]



# ref = [std_model_gams, std_model_list]
# inputmodel = gamsModel(folder, 'mhex_sens_H1.gms')
# outputmodel = solvedModel(folder, 'mhex_sens_H1.lst')

# H1_sens_gams = gamsModel(folder, 'mhex_sens_H1.gms')
# H1_sens_list = solvedModel(folder, 'mhex_sens_H1.lst')

# H2_sens_gams = gamsModel(folder, 'mhex_sens_H2.gms')
# H2_sens_list = solvedModel(folder, 'mhex_sens_H2.lst')

# H3_sens_gams = gamsModel(folder, 'mhex_sens_H3.gms')
# H3_sens_list = solvedModel(folder, 'mhex_sens_H3.lst')

# C1_sens_gams = gamsModel(folder, 'mhex_sens_C1.gms')
# C1_sens_list = solvedModel(folder, 'mhex_sens_C1.lst')

# C2_sens_gams = gamsModel(folder, 'mhex_sens_C2.gms')
# C2_sens_list = solvedModel(folder, 'mhex_sens_C2.lst')

# inputmodels = [H1_sens_gams, H2_sens_gams, H3_sens_gams, C1_sens_gams, C2_sens_gams]
# outputmodels = [H1_sens_list, H2_sens_list, H3_sens_list, C1_sens_list, C2_sens_list]




# print len(std_model_list.Tin)


# names = [	'mhex_sens_H1.lst',	'mhex_sens_H2.lst', \
# 			'mhex_sens_H3.lst',	'mhex_sens_C1.lst', \
# 			'mhex_sens_C2.lst',	'mhex_sens_pN2.lst',\
# 			'mhex_sens_std.lst']

# def mkdir_p(self,path):
#     try:
#         os.makedirs(path)
#     except OSError as exc:
#         if exc.errno == errno.EEXIST:
#             pass
#         else: raise

# def compositeCurve(gamsModel, solvedModel, plotTitle):
# 	Hsqns = ['H1', 'H2', 'H3']	# Hot side
# 	Csqns = ['C1', 'C2']		# Cold side
# 	Htemp = []; Henth = []; Ctemp = []; Centh = []
# 	for spec in Hsqns:
# 		t = float(gamsModel.Tin.get(spec)),float(gamsModel.Tout.get(spec))
# 		dt = t[0] - t[1]
# 		if dt < 0:
# 			dt = dt*-1
# 		h = float(solvedModel.EnthS6)*float(solvedModel.FN2) - dt*float(gamsModel.FCp.get(spec)) \
# 			, float(solvedModel.EnthS6)*float(solvedModel.FN2)
# 		dh = h[0]
# 		Htemp.append(t)
# 		Henth.append(h)
# 	print "Htemp = %s" % sorted(Htemp, reverse = True)
# 	print "Henth = %s" % Henth
# 	# Making intervals
# 	counter = 0
# 	intervals = {} 	# Dictionary on the form ints = {T1:FCp1, T2:FCp2, ...}
# 	while counter+1 <= len(Htemp):
# 		if Htemp[counter][1] < Htemp[counter+1][0]:
# 			pass
# 	'''	Need to check for overlapping temperature intervals, find common FCp and decide enthalpy
# 		values. The enthalpy span is set by the enth-tuples. Make a function that starts the search 
# 		on top temperature for hot and cold streams, respectively.  '''

# def generateHotCurve(model):
# 	sqns = ['H1', 'H2', 'H3']
# 	TCp = []; T = []; H = []; FCp = 0	# T-Cp-dictionary. Positive CP for inlets, negative for outlets.
# 	for spec in sqns:
# 		TCp.append((float(model.Tin.get(spec)),float(model.FCp.get(spec))))
# 		TCp.append((float(model.Tout.get(spec)),-float(model.FCp.get(spec))))
# 	TCp = sorted(TCp, reverse = True)
# 	print TCp
# 	for current, next in pairwise(TCp):
# 		T.append(current[0])
# 		dt = current[0] - next[0]
# 		FCp += current[1]
# 		H.append(FCp*dt)
# 	n = len(TCp)-1
# 	T.append(TCp[n][0])
# 	dt = TCp[n][0] - TCp[n-1][0]
# 	FCp += TCp[n][1]
# 	H.append(FCp*dt)
# 	sumNestedList(H)
# 	H[0] = 0
# 	H = sorted(H, reverse = True)
# 	print "T = %s \n H = %s" % (T,H)
# 	d1 = Gnuplot.Data(H,T,with_="lines")
# 	gp = Gnuplot.Gnuplot()
# 	gp('reset')
# 	gp('set style line 1')
# 	gp('set key horizontal below height 2')
# 	gp.title('Composite Curves')
# 	gp.xlabel('{/Symbol D}H')
# 	gp.ylabel('Temperature [K]')
# 	gp.plot(d1)

# def generateColdCurve(model):
# 	sqns = ['C1', 'C2']
# 	TCp = []; T = []; H = []; FCp = 0
# 	for spec in sqns:
# 		TCp.append((float(model.Tin.get(spec)),float(model.FCp.get(spec))))
# 		TCp.append((float(model.Tout.get(spec)),-float(model.FCp.get(spec))))
# 	TCp = sorted(TCp)
# 	print "TCp = %s" % TCp
# 	for current, next in zip(TCp, TCp[1:]): 
# 		T.append(current[0])
# 		dt = current[0] - next[0]
# 		FCp += current[1]
# 		H.append(FCp*dt)
# 	n = len(TCp)-1
# 	T.append(TCp[n][0])
# 	dt = TCp[n][0] - TCp[n-1][0]
# 	FCp += TCp[n][1]
# 	H.append(FCp*dt)
# 	sumNestedList(H)
# 	H[0] = 0
# 	H = sorted(H)
# 	print "T = %s \n H = %s" % (T,H)
# 	d1 = Gnuplot.Data(H,T,with_="lines")
# 	gp = Gnuplot.Gnuplot()
# 	gp('reset')
# 	gp('set style line 1')
# 	gp('set key horizontal below height 2')
# 	gp.title('Composite Curves')
# 	gp.xlabel('{/Symbol D}H')
# 	gp.ylabel('Temperature [K]')
# 	gp.plot(d1)

# def sumNestedList(thisList):
# 	n = len(thisList)-1
# 	while n > 0:
# 		thisList[n] = sum(thisList[:n])
# 		n -= 1
# 	return thisList

# def pairwise(iterable):
# 	"s -> (s0,s1), (s1,s2), (s2, s3), ..."
# 	a, b = itertools.tee(iterable)
# 	next(b, None)
# 	return itertools.izip(a, b)

# d = [Gnuplot.Data(array(enth[i]),array(temp[i]), with_="lines") for i in range(0,len(temp))]
# gp = Gnuplot.Gnuplot()
# gp('reset')
# gp('set style line 1')
# gp('set key horizontal below height 2')
# gp.title(plotTitle)
# gp.xlabel('{/Symbol D}H')
# gp.ylabel('Temperature [K]')
# gp.plot(d[0], d[1], d[2], d[3], d[4])
# gp.hardcopy('composite_curve.ps', mode = 'eps')

# class HeatFlow:
# 	'''
# 	Class specifying the flow, temperatures and heat capacity of cold and hot flows in the mhex. 
# 	'''
# 	def __init__(self,temp_in,temp_out,FCP, enth_state):
# 		self.Tin = temp_in			# Sets inlet temperature of flow
# 		self.Tout = temp_out		# Sets outlet temperature of flow
# 		self.FCP = FCP				# Sets flow heat capacity
# 		self.Hstate = enth_state	# Reference state of enthalpy 		DO I NEED THIS ONE?

# 	def calcEnthalpy(self,F_N2):	# Isn't this for the optimized N2-flow?
# 		self.enthalpy = [self.Hstate*F_N2-(self.Tin - self.Tout)*self.FCP, self.Hstate*F_N2]

# class HeatFlow:
# 	'''
# 	Class specifying the flow, temperatures and heat capacity of cold and hot flows in the mhex. 
# 	'''
# 	def __init__(self, inputmodel, outputmodel):
# 		'''
# 		Loads the values from a model and assigns them to the flow. 
# 		'''
# 		# self.name = parm 							# Sets name of flow
# 		self.Tin = model.Tin 				# Sets inlet temperature of flow
# 		self.Tout = model.Tout.get(parm)			# Sets outlet temperature of flow
# 		self.FCP = mode.FCP.get(parm)				# Sets flow heat capacity

# class optimizedFlow:
# 	'''
# 	Class for the optimized nitrogen flow, computing whatever needed for the flow. 
# 	'''
# 	def __init__(self,state):				#temp_in,temp_out,FCP,enth_states):
# 		self.Tin = temp_in
# 		self.Tout = temp_out

# def findData2(textlist):
# 	'''
# 	The function extracts values from the file "textlist" and return them as a dictionary on the form:
# 	EnthS2: value \n Tin_C3: value.

# 	NB: should probably change the function to return a nested dictionary, such that all "Tin" values
# 	are in one nest. 
# 	'''
# 	# sqns = ['FN2', 'PresS1', 'PumpWork', 'EnthS2', 'EnthS3', 'EnthS4', 'EnthS5', 'EnthS6', 'Tin', 'Tout', 'FCp', 'Z']
# 	sqns = ['FN2', 'PumpWork', 'Z']
# 	stream = ['C3', 'C4', 'C5', 'C6']
# 	ans = AutoVivification()
# 	flag = False
# 	count = 0
# 	for spec in sqns:
# 		for line in textlist:
# 			if spec == 'Tin' or spec == 'Tout' or spec == 'FCp':
# 				try:
# 					res = re.match('-*\s*VAR '+spec+'\s*$', line)	# Searching for the "---- VAR Tin"
# 				except:
# 					pass
# 				else:
# 					if res != None:
# 						flag = True
# 					if count > 3 and count < 8:
# 						for st in stream:
# 							try:
# 								ans[spec][st] = re.match(st+'\s+\S+\s+([0-9.]+)', line).group(1)
# 							except:
# 								pass
# 							else:
# 								pass
# 					elif count == 8:
# 						flag = False
# 						count = 0
# 					else:
# 						pass
# 			else:
# 				try:
# 					ans[spec] = re.match('-*\s*VAR '+spec+'\s*\S*\s*([-0-9.]*)', line).group(1)
# 				except:
# 					pass
# 			if flag == True:
# 				count += 1
# 	for line in textlist:
# 		try:
# 			ans['Inputfile'] = re.match('Input\s+(\/\S+)+\/(\S+)$', line).group(2)
# 		except:
# 			pass
# 		try:
# 			ans['Outputfile'] = re.match('Output\s+(\/\S+)+\/(\S+)$', line).group(2)
# 		except:
# 			pass
# 	return ans

# class gamsModel:
# 	'''
# 	Class specifying the values fed to the solver from the .gms file. 
# 	'''
# 	def __init__(self, folder, filename):
# 		data = findParameters(readFile(folder, filename), 3, 2)
# 		self.name = filename
# 		self.Tin = data.get('Tin')
# 		self.Tout = data.get('Tout')
# 		self.FCp = data.get('FCp')
# 	def __repr__(self):
# 		return "gamsModel-class"
# 	# def printClass(self,arg):
# 	# 	print self
# 	# 	sortPrintDict(self.arg)
# 	# def printClass(self, arg):
# 	# 	print self.arg
# 	# def __str__(self):
# 		# return sortPrintDict(self)

# class solvedModel:
# 	'''
# 	Class specifying the values retrieved from the .list file of the solved .gms model. 
# 	'''
# 	def __init__(self, folder, filename):
# 		'''
# 		Assigns a name to the model (typically file name), then loads the values from "data".
# 		'''
# 		data = findData(readFile(folder, filename))
# 		self.name = filename
# 		self.inputfile = data.get('Inputfile') 		# Sets name of model
# 		self.outputfile = data.get('Outputfile')
# 		self.FN2 = data.get('FN2')
# 		self.PumpWork = data.get('PumpWork')
# 		self.Z = data.get('Z')
# 		self.Tin = data.get('Tin')
# 		self.Tout = data.get('Tout')
# 		self.FCp = data.get('FCp')
# 		self.EnthS2 = data.get('EnthS2')
# 		self.EnthS3 = data.get('EnthS3')
# 		self.EnthS4 = data.get('EnthS4')
# 		self.EnthS5 = data.get('EnthS5')
# 		self.EnthS6 = data.get('EnthS6')
# 		self.PresS1 = data.get('PresS1')
# 		self.dataDict = data
# 	def printClass(self,arg):
# 		sortPrintDict(self.arg)

# def findData(textlist):
# 	sqns = ['FN2', 'PumpWork', 'PresS1', 'Z', 'FCp', 'Tin']
# 	streams = ['H1', 'H2', 'H3', 'C1', 'C2']
# 	names = ['Input', 'Output']
# 	ans = AutoVivification()
# 	for line in textlist:
# 		for spec in sqns:
# 			try:
# 				ans[spec] = re.match('.*VARIABLE '+spec+'\.L\s*=\s*([0-9.E]+)', line).group(1)
# 			except:
# 				pass
# 			if spec == ('FCp' or 'Tin'):
# 				for Str in streams:
# 					try:
# 						ans[spec][Str] = re.match('.*'+Str+' ([0-9.E]+)', line).group(1)
# 					except:
# 						pass
# 			for name in names:
# 				try:
# 					ans[name+'file'] = re.match(''+name+'\s+(\/\S+)+\/(\S+)$', line).group(2)
# 				except:
# 					pass
# 	return ans

# def readFile(folder, filename):
# 	try:
# 		with open(folder+'/'+filename, 'r') as f:
# 			res = f.read().split('\n')
# 	except IOError:
# 		print "Error: Can't find file or read data named '%s'" % filename
# 	else:
# 		print "File named '%s' from the folder '%s' successfully loaded" % (filename, folder)
# 		f.close()
# 	return res

# class gamsModel:
#     '''
#     Class specifying the values fed to the solver from the .gms file. 
#     '''
#     def __init__(self, folder, filename):
#         data = findParameters(readFile(folder, filename), 3, 2)
#         self.name = filename
#         self.Tin = data.get('Tin')
#         self.Tout = data.get('Tout')
#         self.FCp = data.get('FCp')
#     def __repr__(self):
#         return "gamsModel-class"

# inputmodels = [\
#   gamsModel('sensitivity', 'mhex_sens_H1.gms'), \
#   gamsModel('sensitivity', 'mhex_sens_H2.gms'), \
#   gamsModel('sensitivity', 'mhex_sens_H3.gms'), \
#   gamsModel('sensitivity', 'mhex_sens_C1.gms'), \
#   gamsModel('sensitivity', 'mhex_sens_C2.gms'), \
#   gamsModel('sensitivity', 'mhex_sens_FCP_H1.gms'),\
#   gamsModel('sensitivity', 'mhex_sens_FCP_H2.gms'),\
#   gamsModel('sensitivity', 'mhex_sens_FCP_H3.gms'),\
#   gamsModel('sensitivity', 'mhex_sens_FCP_C1.gms'),\
#   gamsModel('sensitivity', 'mhex_sens_FCP_C2.gms')\
#   ]

# def findParameters(inputlist, hotstreams, coldstreams):
#     '''
#     The function extracts the parameter values from the gams inputfiles, such as temperature
#     goals for the hot and cold streams. The "hotstreams" and "coldstreams" argument is made
#     to make it easy to extend the program to another set of hot and cold streams beside the
#     mhex example: 3 hot and 2 cold streams. 
#     !___The regex has not been adapted to this behaviour, though. ___!
#     '''
#     ans = AutoVivification()
#     sqns = ['FCp', 'Tin', 'Tout']
#     streams = []
#     for i in range(1,hotstreams+1):
#         streams.append('H'+str(i))
#     for i in range(1,coldstreams+1):
#         streams.append('C'+str(i))
#     for line in inputlist:
#         for spec in sqns:
#             for var in streams:
#                 try:
#                     ans[spec][var] = re.match(""+spec+"\.fx\('"+var+"'\) = ([0-9.]*)",line).group(1)
#                 except:
#                     pass
#     return ans

# def grad(modelList,var):
#     '''
#     modelList = [solvedOpt, solvedOpt+, solvedOpt-]
#     var = 'FN2' or 'PumpWork'.
#     h = steplength
#     '''
#     eps = 1.e-6                     # numerical tolerance
#     z_p   = float(modelList[0].Z)   # positive perp
#     z_n   = float(modelList[1].Z)   # negative perp
#     h1 = abs(float(getattr(modelList[1],var)) - float(getattr(ref,var)))
#     h2 = abs(float(getattr(modelList[0],var)) - float(getattr(ref,var)))
#     if abs(h1-h2) > eps:
#         print "h1 != h2"
#     return (z_p - z_n)/(2*h1)

# def Gradient(aModelList):
#     '''
#     modelList = [FN2-opt+/opt-, PW-opt+/opt-]
#     h1 = steplength FN2
#     h2 = steplength PW
#     '''
#     z1 = grad(aModelList[0],'FN2')
#     z2 = grad(aModelList[1],'PumpWork')
#     return [z1,z2]

# def sensitivity(inMlist, outMlist, ref):
#     '''
#     'inMlist' is a list of the gamsModel objects from the .gms-files.
#     'outMlist' is a list of the solvedModel objects from the .lst-files.
#     'ref' is [gamsSTD, solvedSTD], the standard reference gamsModel and solvedModel object.
#     F = (dy_m/dd_n)
#     dy: change in optimal solution, that is, dZ, dFN2, dPumpWork
#     dd: disturbance, that is, temperature change for flows H1, H2, H3, C1, C2.
#     print ref[1].FCp['H3']
#     '''
#     y = ['FN2', 'PumpWork']
#     u = ['H1', 'H2', 'H3', 'C1', 'C2']
#     dy = [[] for _ in range(len(y))]
#     du = []
#     for (u_i,inM) in zip(u,inMlist):
#         try:
#             du.append(float(inM.Tin.get(u_i)) - float(ref[0].Tin.get(u_i)))
#         except:
#             print "ERROR: while reading input %s" % u_i
#     for y_i in y:
#         for outM in outMlist:
#             try:
#                 dy[y.index(y_i)].append(float(getattr(outM, y_i)) - float(getattr(ref[1], y_i)))
#             except AttributeError:
#                 print "ERROR: AttributeError was caught in 'sensitivity'."
#             except:
#                 print "ERROR: in reading output %s" % y_i
#     try:
#         return [[dyj/dui for (dyj,dui) in zip(dyi,du)] for dyi in dy]
#     except ZeroDivisionError:
#         print "Error in F: du_i = 0 for some value of i"
#         return 0
#     except:
#         print "ERROR: some error occured while computing sensitivity."
#         return 0